#include <iostream>

/** Dans ce fichier, on compare le patron de conception Visitor
 * avec le pattern matching.
 * Attention:
 *   - Le patron Visitor, sur un "type de données" récursif, demande
 *     qu'on itère à travers la structure. La combinaison de l'étude
 *     de cas (pattern-matching) et de l'itération fait dès lors
 *     plutôt penser aux opérations fold, mais le Visitor a une
 *     interface plus impérative.
 *   - Le propos de l'analogie est d'aider à comprendre,
 *     pas de comparer les langages / styles de prog en général.
 *     Quitte à regarder plusieurs langages, on peut aller voir
 *     Scala (exemple sur wikipedia: 
 *       http://en.wikipedia.org/wiki/Visitor_pattern) où l'on va
 *     typiquement utiliser match plutôt qu'un Visitor. */

/** Définition du type "visitable"
 * type T = Char of char | Int of int */

class TVisitor;

class T {
  public:
    virtual void accept(TVisitor* v) = 0;
};

class Char : public T {
  public:
    char c;
    Char(char c);
    virtual void accept(TVisitor* v);
};

class Int : public T {
  public:
    int i;
    Int(int i);
    virtual void accept(TVisitor* v);
};

class TVisitor {
  public:
    virtual void visitChar(Char* c)=0;
    virtual void visitInt(Int* i)=0;
};

/** Implémentations des constructeurs et méthodes. */

Char::Char(char c2) : c(c2) {
  // this->c = c2;
}

void Char::accept(TVisitor* v) {
  v->visitChar(this);
}

Int::Int(int i) {
  this->i = i;
}

void Int::accept(TVisitor* v) {
  v->visitInt(this);
}

/** On dérive de TVisitor une classe concrète implémentant une
 * opération possible à effectuer pendant une traversée.
 * Moralement, SomeTVisitor implémente
 *   function
 *     | Char c -> Printf.printf "Un caractère %c\n" c
 *     | Int i -> Printf.printf "Un entier %c\n" c
 */

class SomeTVisitor : public TVisitor {
  public:
    virtual void visitChar(Char* c);
    virtual void visitInt(Int* i);
};

void SomeTVisitor::visitChar(Char* c) {
  std::cout << "Un caractère: " << c->c << std::endl;
}

void SomeTVisitor::visitInt(Int* c) {
  std::cout << "Un entier: " << c->i << std::endl;
}

/** Test de notre visiteur sur deux valeurs. */

int main() {
  (new Char('x'))->accept(new SomeTVisitor);
  (new Int(42))->accept(new SomeTVisitor);
}
